# -*- coding: utf-8 -*-

"""
gaussian_step
A SEAMM plug-in for Gaussian
"""

# Bring up the classes so that they appear to be directly in
# the gaussian_step package.

from .metadata import methods  # noqa: F401
from .metadata import dft_functionals  # noqa: F401
from .metadata import optimization_convergence  # noqa: F401
from .metadata import metadata  # noqa: F401

from .gaussian import Gaussian  # noqa: F401, E501
from .gaussian_parameters import GaussianParameters  # noqa: F401, E501
from .gaussian_step import GaussianStep  # noqa: F401, E501
from .tk_gaussian import TkGaussian  # noqa: F401, E501

from .energy import Energy  # noqa: F401
from .energy_parameters import EnergyParameters  # noqa: F401
from .energy_step import EnergyStep  # noqa: F401
from .tk_energy import TkEnergy  # noqa: F401

from .optimization import Optimization  # noqa: F401
from .optimization_parameters import OptimizationParameters  # noqa: F401
from .optimization_step import OptimizationStep  # noqa: F401
from .tk_optimization import TkOptimization  # noqa: F401

from .thermodynamics_step import ThermodynamicsStep  # noqa: F401
from .thermodynamics import Thermodynamics  # noqa: F401
from .thermodynamics_parameters import ThermodynamicsParameters  # noqa: F401
from .tk_thermodynamics import TkThermodynamics  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = "Paul Saxe"
__email__ = "psaxe@vt.edu"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



# -*- coding: utf-8 -*-

"""Non-graphical part of the Thermodynamics step in a Gaussian flowchart
"""

import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401

import gaussian_step  # noqa: E999
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger("Gaussian")
job = printing.getPrinter()
printer = printing.getPrinter("gaussian")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Thermodynamics(gaussian_step.Optimization):
    """
    The non-graphical part of a Thermodynamics step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : ThermodynamicsParameters
        The control parameters for Thermodynamics.

    See Also
    --------
    TkThermodynamics,
    Thermodynamics, ThermodynamicsParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Thermodynamics",
        extension=None,
        module=__name__,
        logger=logger,
    ):
        """A substep for Thermodynamics in a subflowchart for Gaussian.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Thermodynamics {self}")

        super().__init__(
            flowchart=flowchart,
            title="Thermodynamics",
            extension=extension,
            module=__name__,
            logger=logger,
        )

        self._calculation = "thermodynamics"
        self._model = None
        self._metadata = gaussian_step.metadata
        self.parameters = gaussian_step.ThermodynamicsParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return gaussian_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return gaussian_step.__git_revision__

    def description_text(self, P=None, calculation="Thermodynamics"):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if P is None:
            P = self.parameters.values_to_dict()

        if (
            isinstance(P["optimize first"], bool)
            and P["optimize first"]
            or P["optimize first"] == "yes"
        ):
            calculation = " The structure will be optimized and then the "
        else:
            calculation = "The "

        calculation += "thermodynamic functions will be calculated at {T} and {P}"

        if P["optimize first"]:
            text = gaussian_step.Optimization.description_text(
                self, P=P, calculation=calculation
            )
        else:
            text = gaussian_step.Energy.description_text(
                self, P=P, calculation=calculation
            )

        if (
            isinstance(P["input only"], bool)
            and P["input only"]
            or P["input only"] == "yes"
        ):
            if type(self) is Thermodynamics:
                added = (
                    "\n\nThe input file will be written. No calculation will be run."
                )
                text += "\n" + __(added, **P, indent=4 * " ").__str__()

        return text

    def run(self, keywords=None):
        """Run a Thermodynamics step.

        Parameters
        ----------
        keywords: set()
            Any existing keywords.

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        if keywords is None:
            keywords = set()

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Set the attribute for writing just the input
        self.input_only = P["input only"]

        # Get the current system and configuration (ignoring the system...)
        # _, configuration = self.get_system_configuration(None)

        keywords.add("Freq")
        if P["optimize first"]:
            gaussian_step.Optimization.run(self, keywords=keywords)
        else:
            gaussian_step.Energy.run(self, keywords=keywords)

    def analyze(self, indent="", data={}, out=[], table=None, P=None):
        """Analyze and print the thermodynamic and vibrational data.

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        data: {}
            The data from the Gaussian calculation.
        out: []
            ?
        table: {} = None
            The tabular data for the main output.
        P: {}
            The control parameters.
        """
        if P["optimize first"]:
            gaussian_step.Optimization.analyze(self, data=data, P=P)
        else:
            gaussian_step.Energy.analyze(self, data=data, P=P)



# -*- coding: utf-8 -*-

"""Setup and run Gaussian"""

from collections import Counter
import configparser
import csv
from datetime import datetime, timezone
import gzip
import importlib
import json
import logging
from math import isnan
import os
from pathlib import Path
import pkg_resources
import platform
import pprint
import re
import shutil
import string
import time

import cclib
from cpuinfo import get_cpu_info
import numpy as np
import pandas

import gaussian_step
import seamm
import seamm_exec
import seamm.data
from seamm_util import CompactJSONEncoder, Configuration, Q_
import seamm_util.printing as printing


try:
    from itertools import batched
except ImportError:
    from itertools import islice

    def batched(iterable, n):
        "Batch data into tuples of length n. The last batch may be shorter."
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch


logger = logging.getLogger("Gaussian")
job = printing.getPrinter()
printer = printing.getPrinter("gaussian")


def humanize(memory, suffix="B", kilo=1024):
    """
    Scale memory to its proper format e.g:

        1253656 => '1.20 MiB'
        1253656678 => '1.17 GiB'
    """
    if kilo == 1000:
        units = ["", "k", "M", "G", "T", "P"]
    elif kilo == 1024:
        units = ["", "Ki", "Mi", "Gi", "Ti", "Pi"]
    else:
        raise ValueError("kilo must be 1000 or 1024!")

    for unit in units:
        if memory < 10 * kilo:
            return f"{int(memory)}{unit}{suffix}"
        memory /= kilo


def dehumanize(memory, suffix="B"):
    """
    Unscale memory from its human readable form e.g:

        '1.20 MB' => 1200000
        '1.17 GB' => 1170000000
    """
    units = {
        "": 1,
        "k": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "P": 1000**4,
        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Pi": 1024**4,
    }

    tmp = memory.split()
    if len(tmp) == 1:
        return memory
    elif len(tmp) > 2:
        raise ValueError("Memory must be <number> <units>, e.g. 1.23 GB")

    amount, unit = tmp
    amount = float(amount)

    for prefix in units:
        if prefix + suffix == unit:
            return int(amount * units[prefix])

    raise ValueError(f"Don't recognize the units on '{memory}'")


class Substep(seamm.Node):
    def __init__(
        self,
        flowchart=None,
        title="no title",
        extension=None,
        logger=logger,
        module=__name__,
    ):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(
            flowchart=flowchart, title=title, extension=extension, logger=logger
        )

        self._input_only = False
        self._timing_data = []
        self._timing_path = Path("~/.seamm.d/timing/gaussian.csv").expanduser()

        # Set up the timing information
        self._timing_header = [
            "node",  # 0
            "cpu",  # 1
            "cpu_version",  # 2
            "cpu_count",  # 3
            "cpu_speed",  # 4
            "date",  # 5
            "SMILES",  # 6
            "H_SMILES",  # 7
            "formula",  # 8
            "net_charge",  # 9
            "spin_multiplicity",  # 10
            "model",  # 11
            "keywords",  # 12
            "symmetry",  # 13
            "symmetry_used",  # 14
            "nbf",  # 15
            "nproc",  # 16
            "time",  # 17
        ]
        try:
            self._timing_path.parent.mkdir(parents=True, exist_ok=True)

            self._timing_data = 18 * [""]
            self._timing_data[0] = platform.node()
            tmp = get_cpu_info()
            if "arch" in tmp:
                self._timing_data[1] = tmp["arch"]
            if "cpuinfo_version_string" in tmp:
                self._timing_data[2] = tmp["cpuinfo_version_string"]
            if "count" in tmp:
                self._timing_data[3] = str(tmp["count"])
            if "hz_advertized_friendly" in tmp:
                self._timing_data[3] = tmp["hz_advertized_friendly"]

            if not self._timing_path.exists():
                with self._timing_path.open("w", newline="") as fd:
                    writer = csv.writer(fd)
                    writer.writerow(self._timing_header)
        except Exception:
            self._timing_data = None

    @property
    def version(self):
        """The semantic version of this module."""
        return gaussian_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return gaussian_step.__git_revision__

    @property
    def global_options(self):
        return self.parent.global_options

    @property
    def gversion(self):
        return self.parent.gversion

    @property
    def input_only(self):
        """Whether to write the input only, not run MOPAC."""
        return self._input_only

    @input_only.setter
    def input_only(self, value):
        self._input_only = value

    @property
    def is_runable(self):
        """Indicate whether this not runs or just adds input."""
        return True

    @property
    def options(self):
        return self.parent.options

    def calculate_enthalpy_of_formation(self, data):
        """Calculate the enthalpy of formation from the results of a calculation.

        This uses tabulated values of the enthalpy of formation of the atoms for
        the elements and tabulated energies calculated for atoms with the current
        method.

        Parameters
        ----------
        data : dict
            The results of the calculation.
        """
        if "H" not in data:
            return

        # Read the tabulated values from either user or data directory
        personal_file = Path("~/.seamm.d/data/atom_energies.csv").expanduser()
        if personal_file.exists():
            personal_table = pandas.read_csv(personal_file, index_col=False)
        else:
            personal_table = None

        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        csv_file = path / "atom_energies.csv"
        table = pandas.read_csv(csv_file, index_col=False)

        self.logger.debug(f"self.model = {self.model}")
        self.logger.debug("\n\t".join(table.columns))

        # Check if have the data
        atom_energies = None
        correction_energy = None
        if self.model.startswith("U") or self.model.startswith("R"):
            column = self.model[1:]
        else:
            column = self.model

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"Looking for '{column}' in")
            for i, col in enumerate(table.columns, start=1):
                self.logger.debug(f"\t{i:5d}: {col}")

        column2 = column + " correction"
        if personal_table is not None and column in personal_table.columns:
            atom_energies = personal_table[column].to_list()
            if column2 in personal_table.columns:
                correction_energy = personal_table[column2].to_list()
        elif column in table.columns:
            atom_energies = table[column].to_list()
            if column2 in table.columns:
                correction_energy = table[column2].to_list()

        DfH0gas = None
        if personal_table is not None and "ΔfH°gas" in personal_table.columns:
            DfH0gas = personal_table["ΔfH°gas"].to_list()
        elif "ΔfH°gas" in table.columns:
            DfH0gas = table["ΔfH°gas"].to_list()

        if DfH0gas is None or atom_energies is None:
            return

        # Get the atomic numbers and counts
        _, configuration = self.get_system_configuration(None)
        counts = Counter(configuration.atoms.atomic_numbers)
        Ecorr = 0.0
        for atno, count in counts.items():
            if isnan(atom_energies[atno - 1]):
                # Don't have the data for this element
                return
            Ecorr += count * (DfH0gas[atno - 1] - atom_energies[atno - 1])
            if correction_energy is not None and not isnan(correction_energy[atno - 1]):
                Ecorr += count * correction_energy[atno - 1]

        data["DfH0"] = Q_(data["H"], "E_h").m_as("kJ/mol") + Ecorr

        return

    def cleanup(self):
        """Perform any requested cleanup at the end of the calculation."""
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        handling = P["file handling"]
        if handling == "keep all":
            pass
        elif handling == "remove checkpoint files":
            directory = Path(self.directory)
            step_no = int(self._id[-1])
            chkpoint = directory.parent / f"{step_no}.chk"
            chkpoint.unlink(missing_ok=True)
        elif handling == "remove all":
            directory = Path(self.directory)
            shutil.rmtree(directory)

    def get_functional(self, P=None):
        """Work out the DFT functional"""
        if P is None:
            P = self.parameters.current_values_to_dict(
                context=seamm.flowchart_variables._data
            )
        if P["level"] == "recommended":
            functional = P["functional"]
        else:
            functional = P["advanced_functional"]
        found = functional in gaussian_step.dft_functionals
        if not found:
            # Might be first part of name, or Gaussian-encoded name
            for _key, _data in gaussian_step.dft_functionals.items():
                if functional == _key.split(":")[0].strip():
                    functional = _key
                    found = True
                    break
        if not found:
            # Might be the internal Gaussian name
            for _key, _data in gaussian_step.dft_functionals.items():
                if functional == _data["name"]:
                    functional = _key
                    found = True
                    break
        if not found:
            raise ValueError(f"Don't recognize functional '{functional}'")

        return functional

    def get_method(self, P=None):
        """The method ... HF, DFT, ... used."""
        # Figure out the method.
        if P is None:
            P = self.parameters.current_values_to_dict(
                context=seamm.flowchart_variables._data
            )
        if P["level"] == "recommended":
            method_string = P["method"]
        else:
            method_string = P["advanced_method"]

        # If we don't recognize the string presume (hope?) it is a Gaussian method
        if method_string in gaussian_step.methods:
            method_data = gaussian_step.methods[method_string]
            method = method_data["method"]
        else:
            # See if it matches the keyword part
            for _key, _mdata in gaussian_step.methods.items():
                if method_string == _mdata["method"]:
                    method_string = _key
                    method_data = _mdata
                    method = method_data["method"]
                    break
            else:
                method_data = {}
                method = method_string

        return method, method_data

    def make_plots(self, data):
        """Create the density and orbital plots if requested.

        Parameters
        ----------
        data : dict()
             Dictionary of results from the calculation (results.tag file)
        """
        text = "\n\n"

        directory = Path(self.directory)
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the configuration and basic information
        system, configuration = self.get_system_configuration(None)

        periodicity = configuration.periodicity
        if periodicity != 0:
            raise NotImplementedError("Periodic cube files not implemented yet!")

        # Have the needed data?
        if "homos" not in data:
            return ""

        spin_polarized = len(data["homos"]) == 2

        # Prepare to run
        executor = self.parent.flowchart.executor

        # Read configuration file for Gaussian
        seamm_options = self.global_options
        ini_dir = Path(seamm_options["root"]).expanduser()
        full_config = configparser.ConfigParser()
        full_config.read(ini_dir / "gaussian.ini")
        executor_type = executor.name
        if executor_type not in full_config:
            raise RuntimeError(
                f"No section for '{executor_type}' in MOPAC ini file "
                f"({ini_dir / 'mopac.ini'})"
            )
        config = dict(full_config.items(executor_type))

        # Set up the environment
        if config["root-directory"] != "":
            env = {"g09root": config["root-directory"]}
        else:
            env = {}

        if config["setup-environment"] != "":
            cmd = [". {setup-environment} && cubegen"]
        else:
            cmd = ["cubegen"]

        npts = "-2"

        keys = []
        if P["total density"]:
            keys.append("total density")
        if spin_polarized and P["total spin density"]:
            keys.append("spin density")

        n_errors = 0
        for key in keys:
            if key == "total density":
                args = f"1 Density=SCF gaussian.fchk Total_Density.cube {npts} h"
            elif key == "spin density":
                args = f"1 Spin=SCF gaussian.fchk Spin_Density.cube {npts} h"

            # And run CUBEGEN
            result = executor.run(
                cmd=[*cmd, args],
                config=config,
                directory=self.directory,
                files={},
                return_files=["*"],
                in_situ=True,
                shell=True,
                env=env,
            )
            if not result:
                self.logger.error("There was an error running CubeGen")
                n_errors += 1
                printer.important(f"There was an error calling CUBEGEN, {cmd} {args}")

        # Any requested orbitals
        if P["orbitals"]:
            n_orbitals = data["nmo"]
            # and work out the orbitals
            txt = P["selected orbitals"]
            for spin, homo in enumerate(data["homos"]):
                if txt == "all":
                    orbitals = [*range(n_orbitals)]
                else:
                    orbitals = []
                    for chunk in txt.split(","):
                        chunk = chunk.strip()
                        if ":" in chunk or ".." in chunk:
                            if ":" in chunk:
                                first, last = chunk.split(":")
                            elif ".." in chunk:
                                first, last = chunk.split("..")
                            first = first.strip().upper()
                            last = last.strip().upper()

                            if first == "HOMO":
                                first = homo
                            elif first == "LUMO":
                                first = homo + 1
                            else:
                                first = int(
                                    first.removeprefix("HOMO").removeprefix("LUMO")
                                )
                                if first < 0:
                                    first = homo + first
                                else:
                                    first = homo + 1 + first

                            if last == "HOMO":
                                last = homo
                            elif last == "LUMO":
                                last = homo + 1
                            else:
                                last = int(
                                    last.removeprefix("HOMO").removeprefix("LUMO")
                                )
                                if last < 0:
                                    last = homo + last
                                else:
                                    last = homo + 1 + last

                            orbitals.extend(range(first, last + 1))
                        else:
                            first = chunk.strip().upper()

                            if first == "HOMO":
                                first = homo
                            elif first == "LUMO":
                                first = homo + 1
                            else:
                                first = int(
                                    first.removeprefix("HOMO").removeprefix("LUMO")
                                )
                                if first < 0:
                                    first = homo + first
                                else:
                                    first = homo + 1 + first
                            orbitals.append(first)

                # Remove orbitals out of limits
                tmp = orbitals
                orbitals = []
                for x in tmp:
                    if x >= 0 and x < n_orbitals:
                        orbitals.append(x)

                if spin_polarized:
                    l1 = ("A", "B")[spin]
                    l2 = ("α-", "β-")[spin]
                else:
                    l1 = ""
                    l2 = ""
                for mo in orbitals:
                    if mo == homo:
                        filename = f"{l2}HOMO.cube"
                    elif mo < homo:
                        filename = f"{l2}HOMO-{homo - mo}.cube"
                    elif mo == homo + 1:
                        filename = f"{l2}LUMO.cube"
                    else:
                        filename = f"{l2}LUMO+{mo - homo - 1}.cube"
                    args = f"1 {l1}MO={mo + 1} gaussian.fchk {filename} {npts} h"

                    # And run CUBEGEN
                    result = executor.run(
                        cmd=[*cmd, args],
                        config=config,
                        directory=self.directory,
                        files={},
                        return_files=["*"],
                        in_situ=True,
                        shell=True,
                        env=env,
                    )
                    if not result:
                        self.logger.error("There was an error running CubeGen")
                        n_errors += 1
                        printer.important(
                            f"There was an error calling CUBEGEN, {cmd} {args}"
                        )

        # Finally rename and gzip the cube files
        n_processed = 0
        paths = directory.glob("*.cube")
        for path in paths:
            out = path.with_suffix(".cube.gz")
            with path.open("rb") as f_in:
                with gzip.open(out, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            n_processed += 1
            path.unlink()
        if n_errors > 0:
            text += (
                f"Created {n_processed} density and orbital cube files, but there were "
                f"{n_errors} errors trying to create cube files."
            )
        else:
            text += f"Created {n_processed} density and orbital cube files."

        return text

    def parse_fchk(self, path, data={}):
        """Process the data of a formatted Chk file given as lines of data.

        Parameters
        ----------
        path : pathlib.Path
            The path to the checkpoint file
        """
        lines = path.read_text().splitlines()

        if len(lines) < 2:
            return data

        it = iter(lines)
        # Ignore first potentially truncated title line
        next(it)

        # Type line (A10,A30,A30)
        line = next(it)
        data["calculation"] = line[0:10].strip()
        data["method"] = line[10:40].strip()

        # The rest of the file consists of a line defining the data.
        # If the data is a scalar, it is on the control line, otherwise it follows
        translation = self.metadata["translation"]
        while True:
            try:
                line = next(it)
            except StopIteration:
                break
            try:
                key = line[0:40].strip()
                if key in translation:
                    key = translation[key]
                code = line[43]
                is_array = line[47:49] == "N="
                if is_array:
                    count = int(line[49:61].strip())
                    value = []
                    if code == "I":
                        i = 0
                        while i < count:
                            line = next(it)
                            for pos in range(0, 6 * 12, 12):
                                value.append(int(line[pos : pos + 12].strip()))
                                i += 1
                                if i == count:
                                    break
                    elif code == "R":
                        i = 0
                        while i < count:
                            line = next(it)
                            for pos in range(0, 5 * 16, 16):
                                text = line[pos : pos + 16].strip()
                                # Fortran drops E in format for large exponents...
                                text = re.sub(r"([0-9])-", r"\1E-", text)
                                value.append(float(text))
                                i += 1
                                if i == count:
                                    break
                    elif code == "C":
                        value = ""
                        i = 0
                        while i < count:
                            line = next(it)
                            for pos in range(0, 5 * 12, 12):
                                value += line[pos : pos + 12]
                                i += 1
                                if i == count:
                                    break
                                value = value.rstrip()
                    elif code == "H":
                        value = ""
                        i = 0
                        while i < count:
                            line = next(it)
                            for pos in range(0, 9 * 8, 8):
                                value += line[pos : pos + 8]
                                i += 1
                                if i == count:
                                    break
                                value = value.rstrip()
                    elif code == "L":
                        i = 0
                        while i < count:
                            line = next(it)
                            for pos in range(72):
                                value.append(line[pos] == "T")
                                i += 1
                                if i == count:
                                    break
                else:
                    if code == "I":
                        value = int(line[49:].strip())
                    elif code == "R":
                        value = float(line[49:].strip())
                    elif code == "C":
                        value = line[49:].strip()
                    elif code == "L":
                        value = line[49] == "T"
                data[key] = value
            except Exception:
                pass
        return data

    def parse_output(self, path, data={}):
        """Process the output.

        Parameters
        ----------
        path : pathlib.Path
            The Gaussian log file.
        """
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        lines = path.read_text().splitlines()

        # Did it end properly?
        data["success"] = "Normal termination" in lines[-1]

        # Find the date and version of Gaussian
        # Gaussian 09:  EM64M-G09RevE.01 30-Nov-2015
        it = iter(lines)
        for line in it:
            if "Cite this work" in line:
                for line in it:
                    if "**********************" in line:
                        line = next(it)
                        if "Gaussian" in line:
                            try:
                                _, version, revision, date = line.split()
                                _, month, year = date.split("-")
                                revision = revision.split("Rev")[1]
                                data["G revision"] = revision
                                data["G version"] = f"G{version.strip(':')}"
                                data["G month"] = month
                                data["G year"] = year
                            except Exception as e:
                                self.logger.warning(
                                    f"Could not find the Gaussian citation: {e}"
                                )
                            break
                break

        # And the optimization steps, if any.
        #
        # Need to be careful about end of the first (and presumably only?) optimization.
        # The FORCE calculation prints out the same information about convergence, but
        # may indicate no convergence. This can confuse this code unless we look for the
        # end of the optimization step and quit then
        it = iter(lines)
        n_steps = 0
        max_force = []
        rms_force = []
        max_displacement = []
        rms_displacement = []
        converged = None
        for line in it:
            if line == "         Item               Value     Threshold  Converged?":
                n_steps += 1
                converged = True

                tmp1, tmp2, value, threshold, criterion = next(it).split()
                if tmp1 == "Maximum" and tmp2 == "Force":
                    max_force.append(float(value))
                    data["maximum atom force threshold"] = float(threshold)
                    if criterion != "YES":
                        converged = False

                tmp1, tmp2, value, threshold, criterion = next(it).split()
                if tmp1 == "RMS" and tmp2 == "Force":
                    rms_force.append(float(value))
                    data["RMS atom force threshold"] = float(threshold)
                    if criterion != "YES":
                        converged = False

                tmp1, tmp2, value, threshold, criterion = next(it).split()
                if tmp1 == "Maximum" and tmp2 == "Displacement":
                    max_displacement.append(float(value))
                    data["maximum atom displacement threshold"] = float(threshold)
                    if criterion != "YES":
                        converged = False

                tmp1, tmp2, value, threshold, criterion = next(it).split()
                if tmp1 == "RMS" and tmp2 == "Displacement":
                    rms_displacement.append(float(value))
                    data["RMS atom displacement threshold"] = float(threshold)
                    if criterion != "YES":
                        converged = False
            elif line == " Optimization completed.":
                line = next(it)
                if line == "    -- Stationary point found.":
                    converged = True
                else:
                    self.logger.warning(f"Optimization completed: {line}")
                break
            elif line == "    -- Stationary point found.":
                converged = True
                break

        data["N steps optimization"] = n_steps

        if converged is not None:
            data["optimization is converged"] = converged
            data["maximum atom force"] = max_force[-1]
            data["RMS atom force"] = rms_force[-1]
            data["maximum atom displacement"] = max_displacement[-1]
            data["RMS atom displacement"] = rms_displacement[-1]
        data["maximum atom force trajectory"] = max_force
        data["RMS atom force trajectory"] = rms_force
        data["maximum atom displacement trajectory"] = max_displacement
        data["RMS atom displacement trajectory"] = rms_displacement

        # CBS calculations

        # Complete Basis Set (CBS) Extrapolation:
        # M. R. Nyden and G. A. Petersson, JCP 75, 1843 (1981)
        # G. A. Petersson and M. A. Al-Laham, JCP 94, 6081 (1991)
        # G. A. Petersson, T. Tensfeldt, and J. A. Montgomery, JCP 94, 6091 (1991)
        # J. A. Montgomery, J. W. Ochterski, and G. A. Petersson, JCP 101, 5900 (1994)
        #
        # Temperature=               298.150000 Pressure=                       1.000000
        # E(ZPE)=                      0.050496 E(Thermal)=                     0.053508
        # E(SCF)=                    -78.059017 DE(MP2)=                       -0.281841
        # DE(CBS)=                    -0.071189 DE(MP34)=                      -0.024136
        # DE(Int)=                     0.021229 DE(Empirical)=                 -0.075463
        # CBS-4 (0 K)=               -78.439921 CBS-4 Energy=                 -78.436908
        # CBS-4 Enthalpy=            -78.435964 CBS-4 Free Energy=            -78.460753

        method, method_data = self.get_method(P)
        self.logger.debug(f"Checking for CBS extrapolation for {method}")
        if method[0:4] == "CBS-":
            # Need last section
            if method == "CBS-4M":
                match = "CBS-4 Enthalpy="
            else:
                match = f"{method} Enthalpy="
            self.logger.debug(f"Looking for '{match}'")
            text = []
            found = False
            for line in reversed(lines):
                if found:
                    text.append(line)
                    if "Complete Basis Set" in line:
                        break
                elif match in line:
                    found = True
                    text.append(line)

            self.logger.debug(f"Found CBS extrapolation: {found}")
            if found:
                text = text[::-1]
                it = iter(text)
                next(it)
                citations = []
                for line in it:
                    tmp = line.strip()
                    if tmp == "":
                        break
                    citations.append(tmp)
                data["citations"] = citations

                for line in it:
                    line = line.strip()
                    if len(line) > 40:
                        part = [line[0:37], line[38:]]
                    else:
                        part = [line]
                    for p in part:
                        if "=" not in p:
                            continue
                        key, value = p.split("=", 1)
                        key = key.strip()
                        value = float(value.strip())
                        if "(0 K)" in key:
                            key = "E 0"
                        elif "Free Energy" in key:
                            key = "F"
                        elif "Energy" in key:
                            key = "energy"
                        elif "Enthalpy" in key:
                            key = "H"
                        data[key] = value
                data["model"] = method
                data["Composite/summary"] = "\n".join(text)

        # Gn calculations. No header!!!!!

        # Temperature=              298.150000 Pressure=                      1.000000
        # E(ZPE)=                     0.050251 E(Thermal)=                    0.053306
        # E(CCSD(T))=               -78.321715 E(Empiric)=                   -0.041682
        # DE(Plus)=                  -0.005930 DE(2DF)=                      -0.076980
        # E(Delta-G3XP)=             -0.117567 DE(HF)=                       -0.008255
        # G4(0 K)=                  -78.521880 G4 Energy=                   -78.518825
        # G4 Enthalpy=              -78.517880 G4 Free Energy=              -78.542752

        if method in (
            "G1",
            "G2",
            "G3",
            "G4",
            "G2MP2",
            "G3B3",
            "G3MP2",
            "G3MP2B3",
            "G4MP2",
        ):
            # Need last section
            if method == "G2":
                match = "G2MP2 Enthalpy="
            elif method == "G3B3":
                match = "G3 Enthalpy="
                method = "G3"
            elif method == "G3MP2B3":
                match = "G3MP2 Enthalpy="
                method = "G3MP2"
            else:
                match = f"{method} Enthalpy="
            text = []
            found = False
            for line in reversed(lines):
                if found:
                    if line.strip() == "":
                        break
                    text.append(line)
                elif match in line:
                    found = True
                    text.append(line)

            if found:
                translation = self.metadata["translation"]
                text = text[::-1]
                for line in text:
                    line = line.strip()
                    if len(line) > 36:
                        part = [line[0:36], line[37:]]
                    else:
                        part = [line]
                    for p in part:
                        if "=" not in p:
                            continue
                        key, value = p.split("=", 1)
                        key = key.strip()
                        value = float(value.strip())
                        if key.startswith(method):
                            key = key.split(" ", 1)[1]
                        elif key == "E(Empiric)":
                            key = "E(empirical)"
                        key = "Composite/" + key
                        if key in translation:
                            key = translation[key]
                        data[key] = value

                data["model"] = method
                tmp = " " * 20 + f"{method[0:2]} composite method extrapolation\n\n"
                data["Composite/summary"] = tmp + "\n".join(text)
                data["Total Energy"] = data["Composite/Free Energy"]

        # The Wiberg bond orders ... which look like this:

        # Wiberg bond index matrix in the NAO basis:
        #
        #     Atom    1       2       3       4       5       6       7       8       9
        #     ---- ------  ------  ------  ------  ------  ------  ------  ------  -----
        #   1.  C  0.0000  1.8962  0.0134  0.1327  0.9261  0.9249  0.0071  0.0005  0.00x
        #   2.  C  1.8962  0.0000  1.1131  0.0127  0.0044  0.0049  0.9112  0.0029  0.01x
        #  ...
        #  10.  H  0.0002  0.0022  0.0049  0.9269  0.0000  0.0002  0.0015  0.0171  0.00x
        #
        #     Atom   10
        #     ---- ------
        #   1.  C  0.0002
        #   2.  C  0.0022
        #  ...

        it = iter(lines)
        for line in it:
            n_atoms = None
            if line.startswith(" Wiberg bond index matrix in the NAO basis:"):
                bond_orders = []
                next(it)
                # Read each chunk of output
                while True:
                    # Skip the two header lines
                    next(it)
                    next(it)
                    count = 0
                    # And add the data to the bond_order matrix
                    for line in it:
                        line = line.strip()
                        if line == "":
                            if n_atoms is None:
                                n_atoms = count
                            break
                        count += 1
                        vals = [float(val) for val in line.split()[2:]]
                        if len(bond_orders) < count:
                            bond_orders.append(vals)
                        else:
                            bond_orders[count - 1].extend(vals)
                    if len(bond_orders[0]) >= n_atoms:
                        break

                data["Wiberg bond order matrix"] = bond_orders

        return data

    def parse_punch(self, path, n_atoms, data):
        """Digest the Gaussian punch file.

        Parameters
        ----------
        path : pathlib.Path
            The path to the Punch file.
        n_atoms : int
            The number of atoms in the configuration
        data : dict
            The current data for the calculation.

        Returns
        -------
        dict
        """
        lines = path.read_text().splitlines()
        n_lines = len(lines)

        # Coordinates come first
        if n_lines < n_atoms:
            printer.important("Warning: Punch file does not have coordinates.")
        else:
            XYZs = []
            start = 0
            stop = n_atoms
            for line in lines[start:stop]:
                atno, x, y, z = line.replace("D", "E").split()
                XYZs.append(float(x))
                XYZs.append(float(y))
                XYZs.append(float(z))
            # What is in data[] is the standard orientation not the input orientation
            # so replace with this, which is in the input orientation
            data["Current cartesian coordinates"] = XYZs

        # Gradients next if they are here
        start = stop
        stop = start + n_atoms
        if n_lines > start and n_lines < stop:
            printer.important("Warning: Punch file does not have complete gradients.")
        if n_lines >= 2 * n_atoms:
            XYZs = []
            for line in lines[start:stop]:
                x, y, z = line.replace("D", "E").split()
                XYZs.append(float(x))
                XYZs.append(float(y))
                XYZs.append(float(z))

            if len(XYZs) != 3 * n_atoms:
                raise RuntimeError("Error in the number of gradients in the Punch file")

            if "gradients" in data:
                # Check that the same values to 4 figures
                same = True
                for x0, x1 in zip(data["gradients"], XYZs):
                    if abs(x0 - x1) > 0.001:
                        same = False
                if not same:
                    text = "Warning: gradients from the Punch file differ in\n"
                    text += "\t" + self.header + "\n"
                    for xyz0, xyz1 in zip(
                        batched(data["gradients"], 3), batched(XYZs, 3)
                    ):
                        old = ""
                        new = ""
                        for x0 in xyz0:
                            old += f"{x0:9.4f} "
                        for x1 in xyz1:
                            new += f"{x1:9.4f} "
                        text += f"\n\t{old} <-- {new}"
                    text += "\n\n"
                    printer.important(text)
                data["gradients"] = XYZs

        # Force constants. Triangular matrix, 3 per line.
        ntri = (3 * n_atoms) * (3 * n_atoms + 1) // 2
        n = ntri // 3 if ntri % 3 == 0 else ntri // 3 + 1
        start = stop
        stop = start + n
        if n_lines > start and n_lines < stop:
            printer.important(
                "Warning: Punch file does not have complete force constants."
            )
        if n_lines >= stop:
            count = 0
            d2E = []
            factor = Q_(1.0, "E_h/a_0^2").m_as("kcal/mol/Å^2")
            for line in lines[start:stop]:
                for tmp in line.replace("D", "E").split():
                    d2E.append(float(tmp) * factor)
                    count += 1
                    if count == ntri:
                        break
            if count == ntri:
                data["force constants"] = d2E

        return data

    def process_data(self, data):
        """Massage the cclib data to a more easily used form."""
        self.logger.debug(pprint.pformat(data))
        # Convert numpy arrays to Python lists
        new = {}
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                new[key] = value.tolist()
            elif isinstance(value, list):
                if len(value) > 0 and isinstance(value[0], np.ndarray):
                    new[key] = [i.tolist() for i in value]
                else:
                    new[key] = value
            elif isinstance(value, dict):
                for k, v in value.items():
                    newkey = f"{key}/{k}"
                    if isinstance(v, np.ndarray):
                        new[newkey] = v.tolist()
                    else:
                        new[newkey] = v
            else:
                new[key] = value

        for key in ("metadata/cpu_time", "metadata/wall_time"):
            if key in new:
                time = new[key][0]
                for tmp in new[key][1:]:
                    time += tmp
                new[key] = str(time).lstrip("0:")
                if "." in new[key]:
                    new[key] = new[key].rstrip("0")

        # Pull out the HOMO and LUMO energies as scalars
        if "homos" in new and "moenergies" in new:
            homos = new["homos"]
            if len(homos) == 2:
                for i, letter in enumerate(["alpha", "beta"]):
                    Es = new["moenergies"][i]
                    homo = homos[i]
                    new[f"{letter} HOMO orbital number"] = homo + 1
                    new[f"E {letter} homo"] = Es[homo]
                    if homo > 0:
                        new[f"E {letter} nhomo"] = Es[homo - 1]
                    if homo + 1 < len(Es):
                        new[f"E {letter} lumo"] = Es[homo + 1]
                        new[f"E {letter} gap"] = Es[homo + 1] - Es[homo]
                    if homo + 2 < len(Es):
                        new[f"E {letter} slumo"] = Es[homo + 2]
                    if "mosyms" in new:
                        syms = new["mosyms"][i]
                        new[f"{letter} HOMO symmetry"] = syms[homo]
                        if homo > 0:
                            new[f"{letter} NHOMO symmetry"] = syms[homo - 1]
                        if homo + 1 < len(syms):
                            new[f"{letter} LUMO symmetry)"] = syms[homo + 1]
                        if homo + 2 < len(syms):
                            new[f"{letter} SLUMO symmetry"] = syms[homo + 2]
            else:
                Es = new["moenergies"][0]
                homo = homos[0]
                new["alpha HOMO orbital number"] = homo + 1
                new["beta HOMO orbital number"] = homo + 1
                new["E alpha homo"] = Es[homo]
                new["E beta homo"] = Es[homo]
                if homo > 0:
                    new["E alpha nhomo"] = Es[homo - 1]
                    new["E beta nhomo"] = Es[homo - 1]
                if homo + 1 < len(Es):
                    new["E alpha lumo"] = Es[homo + 1]
                    new["E beta lumo"] = Es[homo + 1]
                    new["E alpha gap"] = Es[homo + 1] - Es[homo]
                    new["E beta gap"] = Es[homo + 1] - Es[homo]
                if homo + 2 < len(Es):
                    new["E alpha slumo"] = Es[homo + 2]
                    new["E beta slumo"] = Es[homo + 2]
                if "mosyms" in new:
                    syms = new["mosyms"][0]
                    new["alpha HOMO symmetry"] = syms[homo]
                    new["beta HOMO symmetry"] = syms[homo]
                    if homo > 0:
                        new["alpha NHOMO symmetry"] = syms[homo - 1]
                        new["beta NHOMO symmetry"] = syms[homo - 1]
                    if homo + 1 < len(syms):
                        new["alpha LUMO symmetry"] = syms[homo + 1]
                        new["beta LUMO symmetry"] = syms[homo + 1]
                    if homo + 2 < len(syms):
                        new["alpha SLUMO symmetry"] = syms[homo + 2]
                        new["beta SLUMO symmetry"] = syms[homo + 2]

        # moments
        if "moments" in new:
            moments = new["moments"]
            new["multipole reference point"] = moments[0]
            new["dipole moment"] = moments[1]
            new["dipole moment magnitude"] = np.linalg.norm(moments[1])
            if len(moments) > 2:
                new["quadrupole moment"] = moments[2]
            if len(moments) > 3:
                new["octapole moment"] = moments[3]
            if len(moments) > 4:
                new["hexadecapole moment"] = moments[4]
            del new["moments"]

        for key in ("symmetry group", "symmetry group used"):
            if key in new:
                new[key] = new[key].capitalize()

        return new

    def run_gaussian(self, keywords, extra_sections={}):
        """Run Gaussian.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Check for successful run, don't rerun
        success_file = directory / "success.dat"
        if not success_file.exists():
            # Get the system & configuration
            system, configuration = self.get_system_configuration(None)

            # Access the options
            options = self.options
            seamm_options = self.global_options

            # Get the computational environment and set limits
            ce = seamm_exec.computational_environment()

            # How many threads to use
            n_cores = ce["NTASKS"]
            self.logger.debug("The number of cores available is {}".format(n_cores))

            if seamm_options["parallelism"] not in ("openmp", "any"):
                n_threads = 1
            else:
                if options["ncores"] == "available":
                    n_threads = n_cores
                else:
                    n_threads = int(options["ncores"])
                if n_threads > n_cores:
                    n_threads = n_cores
                if n_threads < 1:
                    n_threads = 1
                if seamm_options["ncores"] != "available":
                    n_threads = min(n_threads, int(seamm_options["ncores"]))
            ce["NTASKS"] = n_threads
            self.logger.debug(f"Gaussian will use {n_threads} threads.")

            # How much memory to use
            if seamm_options["memory"] == "all":
                mem_limit = ce["MEM_PER_NODE"]
            elif seamm_options["memory"] == "available":
                # For the default, 'available', use in proportion to number of
                # cores used
                mem_limit = ce["MEM_PER_CPU"] * n_threads
            else:
                mem_limit = dehumanize(seamm_options["memory"])

            if options["memory"] == "all":
                memory = ce["MEM_PER_NODE"]
            elif options["memory"] == "available":
                # For the default, 'available', use in proportion to number of
                # cores used
                memory = ce["MEM_PER_CPU"] * n_threads
            else:
                memory = dehumanize(options["memory"])

            memory = min(memory, mem_limit)
            ce["MEM_PER_NODE"] = memory

            # Apply a minimum of 800 MB
            min_memory = dehumanize("800 MB")
            if min_memory > memory:
                memory = min_memory

            # Gaussian allows no decimal points.
            memory = humanize(memory, kilo=1000)

            lines = []
            step_no = int(self._id[-1])
            if step_no > 1:
                last_chkpoint = directory.parent / f"{step_no - 1}.chk"
                if last_chkpoint.exists():
                    lines.append(f"%OldChk={last_chkpoint}")
            chkpoint = directory.parent / f"{step_no}.chk"
            lines.append("%Chk=gaussian.chk")
            lines.append(f"%Mem={memory}")
            lines.append(f"%NProcShared={n_threads}")

            lines.append("# " + " ".join(keywords))
            lines.append("# Punch=(Coord,Derivatives)")

            lines.append(" ")
            lines.append(f"{self.title} of {system.name}/{configuration.name}")
            lines.append(" ")
            lines.append(f"{configuration.charge}    {configuration.spin_multiplicity}")

            # Atoms with coordinates
            symbols = configuration.atoms.symbols
            XYZs = configuration.atoms.coordinates
            for symbol, xyz in zip(symbols, XYZs):
                x, y, z = xyz
                lines.append(f"{symbol:2}   {x:10.6f} {y:10.6f} {z:10.6f}")
            lines.append(" ")

            for section in (
                "Initial force constants",
                "NBO input",
            ):
                if section in extra_sections:
                    lines.extend(extra_sections[section])

            files = {"input.dat": "\n".join(lines)}
            self.logger.debug("input.dat:\n" + files["input.dat"])

            printer.important(
                self.indent + f"    Gaussian will use {n_threads} OpenMP threads and "
                f"up to {memory} of memory.\n"
            )
            if self.input_only:
                # Just write the input files and stop
                for filename in files:
                    path = directory / filename
                    path.write_text(files[filename])
                data = {}
            else:
                executor = self.parent.flowchart.executor

                # Read configuration file for Gaussian if it exists
                executor_type = executor.name
                full_config = configparser.ConfigParser()
                ini_dir = Path(seamm_options["root"]).expanduser()
                path = ini_dir / "gaussian.ini"

                # If the config file doesn't exist, get the default
                if not path.exists():
                    resources = importlib.resources.files("gaussian_step") / "data"
                    ini_text = (resources / "gaussian.ini").read_text()
                    txt_config = Configuration(path)
                    txt_config.from_string(ini_text)
                    txt_config.save()

                full_config.read(ini_dir / "gaussian.ini")

                # Getting desperate! Look for an executable in the path
                if (
                    executor_type not in full_config
                    or "root-directory" not in full_config[executor_type]
                    or "setup-environment" not in full_config[executor_type]
                ):
                    # See if we can find the Gaussian environment variables
                    if "g16root" in os.environ:
                        g_ver = "g16"
                        root_directory = os.environ["g16root"]
                        if "GAUSS_BSDDIR" in os.environ:
                            setup_directory = Path(os.environ["GAUSS_BSDDIR"])
                        else:
                            setup_directory = Path(root_directory) / g_ver / "bsd"
                    elif "g09root" in os.environ:
                        g_ver = "g09"
                        root_directory = os.environ["g09root"]
                        if "GAUSS_BSDDIR" in os.environ:
                            setup_directory = Path(os.environ["GAUSS_BSDDIR"])
                        else:
                            setup_directory = Path(root_directory) / g_ver / "bsd"
                    else:
                        root_directory = None
                        exe_path = shutil.which("g16")
                        if exe_path is None:
                            exe_path = shutil.which("g09")
                        if exe_path is None:
                            raise RuntimeError(
                                f"No section for '{executor_type}' in Gaussian ini file"
                                f" ({ini_dir / 'gaussian.ini'}), nor in the defaults, "
                                "nor in the path!"
                            )
                        g_ver = exe_path.name
                        root_directory = str(exe_path.parent.parent)
                        setup_directory = Path(root_directory) / g_ver / "bsd"
                    setup_environment = str(setup_directory / f"{g_ver}.profile")

                    txt_config = Configuration(path)

                    if not txt_config.section_exists(executor_type):
                        txt_config.add_section(executor_type)

                    txt_config.set_value(executor_type, "installation", "local")
                    txt_config.set_value(executor_type, "code", g_ver)
                    txt_config.set_value(
                        executor_type, "root-directory", root_directory
                    )
                    txt_config.set_value(
                        executor_type, "setup-environment", setup_environment
                    )
                    txt_config.save()
                    full_config.read(ini_dir / "gaussian.ini")

                config = dict(full_config.items(executor_type))
                # Use the matching version of the seamm-gaussian image by default.
                config["version"] = self.version

                g_ver = config["code"]

                # Setup the calculation environment definition
                if config["root-directory"] != "":
                    env = {f"{g_ver}root": config["root-directory"]}
                else:
                    env = {}

                if config["setup-environment"] != "":
                    cmd = f". {config['setup-environment']} ; {g_ver}"
                else:
                    cmd = g_ver

                if "scratch-dir" in config and config["scratch-dir"] != "":
                    path = Path(config["scratch-dir"])
                    try:
                        path.mkdir(parents=True, exist_ok=True)
                    except Exception:
                        pass
                    else:
                        env["GAUSS_SCRDIR"] = config["scratch-dir"]

                cmd += " < input.dat > output.txt && formchk gaussian.chk"

                return_files = [
                    "output.txt",
                    "gaussian.chk",
                    "gaussian.fchk",
                    "fort.7",
                ]

                self.logger.debug(f"{cmd=}")
                self.logger.debug(f"{env=}")

                self._timing_data[12] = " ".join(keywords)
                self._timing_data[5] = datetime.now(timezone.utc).isoformat()
                t0 = time.time_ns()

                result = executor.run(
                    cmd=[cmd],
                    config=config,
                    directory=self.directory,
                    files=files,
                    return_files=return_files,
                    in_situ=True,
                    shell=True,
                    env=env,
                )

                t = (time.time_ns() - t0) / 1.0e9
                if self._timing_data is not None:
                    self._timing_data[17] = f"{t:.3f}"
                    self._timing_data[16] = str(n_threads)

                # Check for errors
                chkpoint_ok = True
                if "stderr" in result and result["stderr"] != "":
                    if "formchk" in result["stderr"]:
                        # Something happened with formchk, but it is not fatal
                        # However, don't reuse the checkpoint file
                        chkpoint_ok = False
                    printer.normal("\n\nThere was an error running Gaussian:")
                    for line in result["stderr"].split("\n"):
                        printer.normal("\t" + line)
                if not result:
                    self.logger.error("There was an error running Gaussian")
                    return None

                if chkpoint_ok:
                    (directory / "gaussian.chk").rename(chkpoint)
                else:
                    (directory / "gaussian.chk").unlink(missing_ok=True)

        if not self.input_only:
            # Reget or parse the data
            data_file = directory / "data.json"
            if data_file.exists():
                with data_file.open("rt") as fd:
                    data = json.load(fd)
                self.model = data["model"][9:]
            else:
                # And output
                path = directory / "output.txt"
                if path.exists():
                    try:
                        data = vars(cclib.io.ccread(path))
                        data = self.process_data(data)
                    except Exception as e:
                        self.logger.warning(
                            f"\ncclib raised exception {e}\nPlease report!"
                        )
                        data = {}
                else:
                    data = {}

                # Switch to standard names
                translation = self.metadata["translation"]
                keys = [*data.keys()]
                for key in keys:
                    if key in translation:
                        to = translation[key]
                        if to in data and data[to] != data[key]:
                            self.logger.warning(
                                f"Overwriting {to} with {key}\n"
                                f"\t{data[key]} --> {data[to]}"
                            )
                        data[to] = data[key]
                        del data[key]

                # Adding timing information from SEAMM
                data["SEAMM elapsed time"] = round(t, 1)
                data["SEAMM np"] = n_threads

                self.logger.debug("after cclib")
                self.logger.debug(f"{pprint.pformat(data)}")
                self.logger.debug(80 * "*")

                success = "success" if "success" in data else False

                if not success:
                    raise RuntimeError("Gaussian did not complete successfully")

                # Get the data from the formatted checkpoint file
                data = self.parse_fchk(directory / "gaussian.fchk", data)

                # Debug output
                if self.logger.isEnabledFor(logging.DEBUG):
                    keys = "\n".join(data.keys())
                    self.logger.debug("After parse_fchk")
                    self.logger.debug(f"Data keys:\n{keys}")
                    self.logger.debug(f"Data:\n{pprint.pformat(data)}")

                # And parse a bit more out of the output
                if path.exists():
                    data = self.parse_output(path, data)

                # And the Punch file, if it exists
                punch = Path(directory / "fort.7")
                if punch.exists():
                    data = self.parse_punch(punch, configuration.n_atoms, data)

                # The model chemistry
                if "model" in data:
                    self.model = data["model"]
                elif "method" in data:
                    if data["method"].endswith("DFT"):
                        model = data["method"] + "/" + data["density functional"]
                    else:
                        model = data["method"]

                    # Check if the method uses a basis set
                    methods = gaussian_step.methods
                    tmp = data["method"][1:]
                    tmp = [m for m, d in methods.items() if d["method"] == tmp]
                    if (
                        len(tmp) == 1
                        and "nobasis" in methods[tmp[0]]
                        and methods[tmp[0]]["nobasis"]
                    ):
                        self.model = model
                    elif "basis set name" in data:
                        self.model = model + "/" + data["basis set name"]
                    else:
                        self.model = model
                else:
                    self.model = "unknown"
                self.logger.debug(f"model = {self.model}")
                data["model"] = "Gaussian/" + self.model

                # Capitalize symmetry names
                for key in ("symmetry group", "symmetry group used"):
                    if key in data:
                        data[key] = data[key].capitalize()

                # Check for QCI and change the energy names
                if "QCISD" in data["model"]:
                    if "E cc" in data:
                        data["E qcisd"] = data["E cc"]
                        del data["E cc"]
                    if "E ccsd_t" in data:
                        data["E qcisd_t"] = data["E ccsd_t"]
                        del data["E ccsd_t"]

                # printer.normal(f"\n\n\n\n\nData:\n{pprint.pformat(data)}\n\n\n\n\n")

                # Debug output
                if self.logger.isEnabledFor(logging.DEBUG):
                    keys = "\n".join(data.keys())
                    self.logger.debug(f"Data keys:\n{keys}")
                    self.logger.debug(f"Data:\n{pprint.pformat(data)}")

                # Save the parsed data as JSON
                with data_file.open("w") as fd:
                    json.dump(
                        data, fd, indent=4, sort_keys=True, cls=CompactJSONEncoder
                    )

            # If ran successfully, put out the success file
            if data["success"]:
                success_file.write_text("success")

                if self._timing_data is not None:
                    self._timing_data[11] = data["model"]
                    if "metadata/symmetry_detected" in data:
                        self._timing_data[13] = str(data["metadata/symmetry_detected"])
                    else:
                        self._timing_data[13] = ""
                    if "metadata/symmetry_used" in data:
                        self._timing_data[14] = str(data["metadata/symmetry_used"])
                    else:
                        self._timing_data[14] = ""
                    if "Number of basis functions" in data:
                        self._timing_data[15] = str(data["Number of basis functions"])
                    else:
                        self._timing_data[15] = ""
                    try:
                        with self._timing_path.open("a", newline="") as fd:
                            writer = csv.writer(fd)
                            writer.writerow(self._timing_data)
                    except Exception:
                        pass

        # Add other citations here or in the appropriate place in the code.
        # Add the bibtex to data/references.bib, and add a self.reference.cite
        # similar to the above to actually add the citation to the references.
        if "G version" in data:
            try:
                template = string.Template(self._bibliography[data["G version"]])
                citation = template.substitute(
                    month=data["G month"],
                    version=data["G revision"],
                    year=data["G year"],
                )
                self.references.cite(
                    raw=citation,
                    alias="Gaussian",
                    module="gaussian_step",
                    level=1,
                    note="The principle Gaussian citation.",
                )
            except Exception:
                pass

        return data



# -*- coding: utf-8 -*-

"""Non-graphical part of the Gaussian step in a SEAMM flowchart
"""

import logging
import pkg_resources
from pathlib import Path
import pprint  # noqa: F401
import sys

import gaussian_step
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger("Gaussian")
job = printing.getPrinter()
printer = printing.getPrinter("Gaussian")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


def humanize(memory, suffix="B", kilo=1024):
    """
    Scale memory to its proper format e.g:

        1253656 => '1.20 MiB'
        1253656678 => '1.17 GiB'
    """
    if kilo == 1000:
        units = ["", "k", "M", "G", "T", "P"]
    elif kilo == 1024:
        units = ["", "Ki", "Mi", "Gi", "Ti", "Pi"]
    else:
        raise ValueError("kilo must be 1000 or 1024!")

    for unit in units:
        if memory < 10 * kilo:
            return f"{int(memory)}{unit}{suffix}"
        memory /= kilo


def dehumanize(memory, suffix="B"):
    """
    Unscale memory from its human readable form e.g:

        '1.20 MB' => 1200000
        '1.17 GB' => 1170000000
    """
    units = {
        "": 1,
        "k": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "P": 1000**4,
        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Pi": 1024**4,
    }

    tmp = memory.split()
    if len(tmp) == 1:
        return memory
    elif len(tmp) > 2:
        raise ValueError("Memory must be <number> <units>, e.g. 1.23 GB")

    amount, unit = tmp
    amount = float(amount)

    for prefix in units:
        if prefix + suffix == unit:
            return int(amount * units[prefix])

    raise ValueError(f"Don't recognize the units on '{memory}'")


class Gaussian(seamm.Node):
    """
    The non-graphical part of a Gaussian step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : GaussianParameters
        The control parameters for Gaussian.

    See Also
    --------
    TkGaussian,
    Gaussian, GaussianParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Gaussian",
        namespace="org.molssi.seamm.gaussian",
        extension=None,
        logger=logger,
    ):
        """A step for Gaussian in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        namespace : str
            The namespace for the plug-ins of the subflowchart
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Gaussian {self}")
        self.subflowchart = seamm.Flowchart(
            parent=self, name="Gaussian", namespace=namespace
        )

        super().__init__(
            flowchart=flowchart,
            title="Gaussian",
            extension=extension,
            module=__name__,
            logger=logger,
        )

        self.parameters = gaussian_step.GaussianParameters()
        self._gversion = "g09"
        self._data = {}

    @property
    def version(self):
        """The semantic version of this module."""
        return gaussian_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return gaussian_step.__git_revision__

    @property
    def gversion(self):
        """The Gaussian version to target."""
        return self._gversion

    @gversion.setter
    def gversion(self, version):
        self._gversion = version

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()

    def create_parser(self):
        """Setup the command-line / config file parser"""
        # parser_name = 'gaussian-step'
        parser_name = self.step_type
        parser = self.flowchart.parser

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        # Options for Gaussian
        parser.add_argument(
            parser_name,
            "--max-atoms-to-print",
            default=25,
            help="Maximum number of atoms to print charges, etc.",
        )

        parser.add_argument(
            parser_name,
            "--gaussian-path",
            default="",
            help="the path to the Gaussian executable",
        )

        parser.add_argument(
            parser_name,
            "--gaussian-exe",
            default="g16",
            help="the Gaussian executable",
        )

        parser.add_argument(
            parser_name,
            "--gaussian-root",
            default="",
            help="The location of the root direction for the Gaussian installation",
        )

        parser.add_argument(
            parser_name,
            "--gaussian-environment",
            default="",
            help="A file to source to setup the Gaussian environment",
        )

        parser.add_argument(
            parser_name,
            "--ncores",
            default="4",
            help="How many threads to use in Gaussian",
        )

        parser.add_argument(
            parser_name,
            "--memory",
            default="available",
            help=(
                "The maximum amount of memory to use for Gaussian, which can be "
                "'all' or 'available', or a number, which may use k, Ki, "
                "M, Mi, etc. suffixes. Default: available."
            ),
        )

        return result

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            try:
                text += __(node.description_text(), indent=3 * " ").__str__()
            except Exception as e:
                print(f"Error describing gaussian flowchart: {e} in {node}")
                logger.critical(f"Error describing gaussian flowchart: {e} in {node}")
                raise
            except:  # noqa: E722
                print(
                    "Unexpected error describing gaussian flowchart: {} in {}".format(
                        sys.exc_info()[0], str(node)
                    )
                )
                logger.critical(
                    "Unexpected error describing gaussian flowchart: {} in {}".format(
                        sys.exc_info()[0], str(node)
                    )
                )
                raise
            text += "\n"
            node = node.next()

        return text

    def run(self):
        """Run a Gaussian step.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # The node after this one, to return at end
        next_node = super().run(printer)

        printer.important(self.header)
        printer.important("")

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        while node is not None:
            if node.is_runable:
                node.run()
            node = node.next()

        # Handle any cleanup of files requested
        node = self.subflowchart.get_node("1").next()
        while node is not None:
            if node.is_runable:
                node.cleanup()
            node = node.next()

        return next_node

    def analyze(self, indent="", fchk=[], output=[], configuration=None, **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        # Loop over the subnodes, asking them to do their analysis
        while node is not None:
            for value in node.description:
                printer.important(value)
            node.analyze(data=self._data, configuration=configuration)
            printer.normal("")
            node = node.next()

        # Update the structure
        if "Current cartesian coordinates" in self._data:
            factor = Q_(1, "a0").to("Å").magnitude
            system_db = self.get_variable("_system_db")
            configuration = system_db.system.configuration
            xs = []
            ys = []
            zs = []
            it = iter(self._data["Current cartesian coordinates"])
            for x in it:
                xs.append(factor * x)
                ys.append(factor * next(it))
                zs.append(factor * next(it))
            configuration.atoms["x"][0:] = xs
            configuration.atoms["y"][0:] = ys
            configuration.atoms["z"][0:] = zs
            printer.important(
                self.indent + "    Updated the system with the structure from Gaussian",
            )
            printer.important("")




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2024-12-11T17:03:00-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "b2c8d3f1c59ea8273e29b0929b451f3d4b44c0ab",
 "version": "2024.12.11"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# -*- coding: utf-8 -*-
"""Global control parameters for Gaussian
"""

import logging

from gaussian_step import methods, dft_functionals
import seamm

logger = logging.getLogger("Gaussian")


class EnergyParameters(seamm.Parameters):
    """The control parameters for the energy."""

    parameters = {
        "input only": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "yes",
                "no",
            ),
            "format_string": "s",
            "description": "Write the input files and stop:",
            "help_text": "Don't run MOPAC. Just write the input files.",
        },
        "level": {
            "default": "recommended",
            "kind": "string",
            "format_string": "s",
            "enumeration": ("recommended", "advanced"),
            "description": "The level of disclosure in the interface",
            "help_text": (
                "How much detail to show in the GUI. Currently 'recommended' "
                "or 'advanced', which shows everything."
            ),
        },
        "basis": {
            "default": "6-31G**",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "6-31G",
                "6-31G*",
                "6-31G**",
                "cc-pVDZ",
                "cc-pVTZ",
                "cc-pVQZ",
                "Def2SV",
                "Def2SVP",
                "Def2SVPP",
                "Def2TZP",
            ),
            "format_string": "s",
            "description": "Basis:",
            "help_text": ("The basis set to use."),
        },
        "method": {
            "default": "DFT: Kohn-Sham density functional theory",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in methods if methods[x]["level"] == "normal"],
            "format_string": "s",
            "description": "Method:",
            "help_text": ("The computational method to use."),
        },
        "advanced_method": {
            "default": "DFT: Kohn-Sham density functional theory",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in methods],
            "format_string": "s",
            "description": "Method:",
            "help_text": ("The computational method to use."),
        },
        "functional": {
            "default": "B3LYP : hybrid functional of Becke and Lee, Yang, and Parr",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [
                x for x in dft_functionals if dft_functionals[x]["level"] == "normal"
            ],
            "format_string": "s",
            "description": "DFT Functional:",
            "help_text": ("The exchange-correlation functional to use."),
        },
        "advanced_functional": {
            "default": "B3LYP : hybrid functional of Becke and Lee, Yang, and Parr",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in dft_functionals],
            "format_string": "s",
            "description": "DFT Functional:",
            "help_text": ("The exchange-correlation functional to use."),
        },
        "dispersion": {
            "default": "GD3BJ",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ["none", "GD3BJ", "GD3", "DG2"],
            "format_string": "s",
            "description": "Dispersion correction:",
            "help_text": ("The dispersion correction to use."),
        },
        "spin-restricted": {
            "default": "default",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("default", "yes", "no"),
            "format_string": "s",
            "description": "Spin-restricted:",
            "help_text": (
                "Whether to restrict the spin (RHF, ROHF, RKS) or not "
                "(UHF, UKS)."
                " Default is restricted for singlets, unrestricted otherwise."
            ),
        },
        "convergence": {
            "default": "default",
            "kind": "integer",
            "default_units": "",
            "enumeration": ("default",),
            "format_string": "s",
            "description": "Energy convergence criterion:",
            "help_text": (
                "Criterion for convergence of the RMS of the density (10^-N) and "
                "maximum change in the density matrix (10^-(N+2))."
            ),
        },
        "maximum iterations": {
            "default": "default",
            "kind": "integer",
            "default_units": "",
            "enumeration": ("default",),
            "format_string": "s",
            "description": "Maximum iterations:",
            "help_text": "Maximum number of SCF iterations.",
        },
        "ignore convergence": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Ignore lack of convergence:",
            "help_text": (
                "Whether to ignore lack of convergence in the SCF. Otherwise, "
                "an error is thrown."
            ),
        },
        "use symmetry": {
            "default": "yes",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("yes", "loose", "identify only", "no"),
            "format_string": "s",
            "description": "Use symmetry:",
            "help_text": "Whether to use symmetry, and if so how much.",
        },
        "calculate gradient": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Calculate gradient:",
            "help_text": "Whether to calculate the gradient:",
        },
        "freeze-cores": {
            "default": "yes",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Freeze core orbitals:",
            "help_text": (
                "Whether to freeze the core orbitals in correlated " "methods"
            ),
        },
        "bond orders": {
            "default": "Wiberg",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("Wiberg", "none"),
            "format_string": "s",
            "description": "Calculate bond orders:",
            "help_text": "Whether to calculate the bond orders and if so how.",
        },
        "apply bond orders": {
            "default": "yes",
            "kind": "bool",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Apply bond orders to structure:",
            "help_text": (
                "Whether to use the calculated bond orders to update the structure"
            ),
        },
        "file handling": {
            "default": "remove checkpoint files",
            "kind": "string",
            "format_string": "s",
            "enumeration": ("keep all", "remove all", "remove checkpoint files"),
            "description": "File handling",
            "help_text": (
                "How to handle files after a successful calculation."
                "or 'advanced', which shows everything."
            ),
        },
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": ("The results to save to variables or in " "tables. "),
        },
        "create tables": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Create tables as needed:",
            "help_text": (
                "Whether to create tables as needed for "
                "results being saved into tables."
            ),
        },
    }

    output_parameters = {
        "total density": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot total density:",
            "help_text": "Whether to plot the total charge density.",
        },
        "total spin density": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot total spin density:",
            "help_text": "Whether to plot the total spin density.",
        },
        "difference density": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot difference density:",
            "help_text": "Whether to plot the difference density.",
        },
        "orbitals": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot orbitals:",
            "help_text": "Whether to plot orbitals.",
        },
        "selected orbitals": {
            "default": "HOMO, LUMO",
            "kind": "string",
            "default_units": "",
            "enumeration": ("HOMO, LUMO", "-1, HOMO, LUMO, +1", "all"),
            "format_string": "",
            "description": "Selected orbitals:",
            "help_text": "Which orbitals to plot.",
        },
        "region": {
            "default": "default",
            "kind": "string",
            "default_units": "",
            "enumeration": ("default", "explicit"),
            "format_string": "",
            "description": "Region:",
            "help_text": "The region for the plots",
        },
        "nx": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "Grid:",
            "help_text": "Number of grid points in first direction",
        },
        "ny": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "x",
            "help_text": "Number of grid points in second direction",
        },
        "nz": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "x",
            "help_text": "Number of grid points in first direction",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **EnergyParameters.parameters,
                **EnergyParameters.output_parameters,
                **seamm.standard_parameters.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )

        # Do any local editing of defaults
        tmp = self["configuration name"]
        tmp._data["enumeration"] = ["single-point with {model}", *tmp.enumeration[1:]]
        tmp.default = "keep current name"

        tmp = self["configuration name"]
        tmp._data["enumeration"] = ["single-point with {model}", *tmp.enumeration]
        tmp.default = "single-point with {model}"



"""This file contains metadata describing the results from Gaussian
"""

metadata = {}

methods = {
    "AM1: AM1 semiempirical HF": {
        "method": "AM1",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PM3: PM3 semiempirical HF": {
        "method": "PM3",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PM3MM: PM3 semiempirical HF with MM corrections": {
        "method": "PM3MM",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PM6: PM6 semiempirical HF": {
        "method": "PM6",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PDDG: PDDG variant of PM6 semiempirical HF": {
        "method": "PDDG",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PM7: modified PM7 semiempirical HF": {
        "method": "PM7",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "PM7MOPAC: PM7 semiempirical HF as in MOPAC": {
        "method": "PM7MOPAC",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "nobasis": True,
    },
    "DFT: Kohn-Sham density functional theory": {
        "method": "DFT",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
    },
    "HF: Hartree-Fock self consistent field (SCF)": {
        "method": "HF",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
    },
    "MP2: 2nd-order Møller–Plesset perturbation theory": {
        "method": "MP2",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "MP3: 3rd-order Møller–Plesset perturbation theory": {
        "method": "MP3",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "MP4(SDTQ): 4th-order Møller–Plesset perturbation theory": {
        "method": "MP4",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "numeric",
        "freeze core?": True,
    },
    "MP4(SDQ): 4th-order Møller–Plesset perturbation theory": {
        "method": "MP4(SDQ)",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "MP5: 5th-order Møller–Plesset perturbation theory": {
        "method": "MP5",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "QCISD: Quadratic CI": {
        "method": "QCI",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "QCISD(T): Quadratic CI with triples correction": {
        "method": "QCISD(T)",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "numerical",
        "freeze core?": True,
    },
    "CCD: coupled cluster doubles": {
        "method": "CCD",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "analytic",
        "freeze core?": True,
    },
    "CCSD: coupled cluster singles & doubles": {
        "method": "CCSD",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "numeric",
        "freeze core?": True,
    },
    "CCSD(T): coupled cluster singles & doubles plus triples": {
        "method": "CCSD(T)",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "numeric",
        "freeze core?": True,
    },
    "CBS-4M: Complete Basis Set method of Petersson, et al.": {
        "method": "CBS-4M",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "CBS-QB3: Complete Basis Set method of Petersson, et al.": {
        "method": "CBS-QB3",
        "calculation": ["energy", "gradients"],
        "level": "normal",
        "gradients": "none",
        "nobasis": True,
    },
    "CBS-APNO: Complete Basis Set method of Petersson, et al.": {
        "method": "CBS-APNO",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G1: Gaussian-1 composite method": {
        "method": "G1",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G2: Gaussian-2 composite method": {
        "method": "G2",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G2MP2: Gaussian-2 MP2 composite method": {
        "method": "G2MP2",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G3: Gaussian-3 composite method": {
        "method": "G3",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G3MP2: Gaussian-3 MP2 composite method": {
        "method": "G3MP2",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G3B3: Gaussian-3 B3LYP composite method": {
        "method": "G3B3",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G3MP2B3: Gaussian-3 MP2 B3LYP composite method": {
        "method": "G3MP2B3",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G4: Gaussian-4 composite method": {
        "method": "G4",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
    "G4MP2: Gaussian-4 MP2 composite method": {
        "method": "G4MP2",
        "calculation": ["energy", "gradients"],
        "level": "advanced",
        "gradients": "none",
        "nobasis": True,
    },
}

dft_functionals = {
    "PBE : GGA of Perdew, Burke and Ernzerhof": {
        "name": "PBEPBE",
        "dispersion": ["none", "GD3BJ", "GD3", "GD2", "PFD"],
        "level": "normal",
    },
    "PBE-98 : GGA of Perdew, Burke and Ernzerhof, revised": {
        "name": "PBEhPBE",
        "dispersion": ["none", "PFD"],
        "level": "normal",
    },
    "B3LYP : hybrid functional of Becke and Lee, Yang, and Parr": {
        "name": "B3LYP",
        "dispersion": ["none", "GD3BJ", "GD3", "GD2", "PFD"],
        "level": "normal",
    },
    "HSE06 : hybrid functional of Heyd, Scuseria and Ernzerhof": {
        "name": "HSEH1PBE",
        "dispersion": ["none", "PFD"],
        "level": "normal",
    },
    "PBE0 : hybrid functional of Perdew, Burke and Ernzerhof, and Adamo": {
        "name": "PBE1PBE",
        "dispersion": ["none", "GD3BJ", "GD3", "PFD"],
        "level": "normal",
    },
}

optimization_convergence = {
    "default": "",
    "tight": "Tight",
    "very tight": "VeryTight",
    "loose": "Loose",
}

"""Properties that Gaussian produces.
`metadata["results"]` describes the results that this step can produce. It is a
dictionary where the keys are the internal names of the results within this step, and
the values are a dictionary describing the result. For example::

    metadata["results"] = {
        "total_energy": {
            "calculation": [
                "energy",
                "optimization",
            ],
            "description": "The total energy",
            "dimensionality": "scalar",
            "methods": [
                "ccsd",
                "ccsd(t)",
                "dft",
                "hf",
            ],
            "property": "total energy#QuickMin#{model}",
            "type": "float",
            "units": "E_h",
        },
    }

Fields
______

calculation : [str]
    Optional metadata describing what subtype of the step produces this result.
    The subtypes are completely arbitrary, but often they are types of calculations
    which is why this is name `calculation`. To use this, the step or a substep
    define `self._calculation` as a value. That value is used to select only the
    results with that value in this field.

description : str
    A human-readable description of the result.

dimensionality : str
    The dimensions of the data. The value can be "scalar" or an array definition
    of the form "[dim1, dim2,...]". Symmetric tringular matrices are denoted
    "triangular[n,n]". The dimensions can be integers, other scalar
    results, or standard parameters such as `n_atoms`. For example, '[3]',
    [3, n_atoms], or "triangular[n_aos, n_aos]".

methods : str
    Optional metadata like the `calculation` data. `methods` provides a second
    level of filtering, often used for the Hamiltionian for *ab initio* calculations
    where some properties may or may not be calculated depending on the type of
    theory.

property : str
    An optional definition of the property for storing this result. Must be one of
    the standard properties defined either in SEAMM or in this steps property
    metadata in `data/properties.csv`.

type : str
    The type of the data: string, integer, or float.

units : str
    Optional units for the result. If present, the value should be in these units.
"""
composite = (
    "CBS-4M",
    "CBS-QB3",
    "CBS-APNO",
    "G1",
    "G2",
    "G3",
    "G4",
    "G2MP2",
    "G3B3",
    "G3MP2",
    "G3MP2B3",
    "G4MP2",
)

metadata["translation"] = {
    "Alpha Orbital Energies": "alpha orbital energies",
    "Cartesian Force Constants": "force constants",
    "Cartesian Gradient": "gradients",
    "Cluster Energy with triples": "E ccsd_t",
    "Cluster Energy": "E cc",
    "Composite/DE(2DF)": "delta E 2df",
    "Composite/DE(CBS)": "deltaE cbs",
    "Composite/DE(CCSD)": "delta E ccsd",
    "Composite/DE(HF)": "delta E hf",
    "Composite/DE(MP2)": "deltaE mp2",
    "Composite/DE(MP34)": "deltaE mp34",
    "Composite/DE(Plus)": "delta E diffuse",
    "Composite/DE(empirical)": "delta E empirical",
    "Composite/DE(int)": "delta E internals",
    "Composite/E(0 K)": "E 0",
    "Composite/E(CCSD(T))": "E ccsd_t",
    "Composite/E(Delta-G3XP)": "delta E g3xp",
    "Composite/E(SCF)": "E scf",
    "Composite/E(Thermal)": "E thermal",
    "Composite/E(ZPE)": "ZPE",
    "Composite/Energy": "energy",
    "Composite/Enthalpy": "H",
    "Composite/Free Energy": "F",
    "Composite/Pressure": "P",
    "Composite/Temperature": "T",
    "Dipole Derivatives": "dipole derivatives",
    "Gaussian Version": "code version",
    "Geometry Optimization Converged": "optimization is converged",
    "HyperPolarizability": "hyperpolarizability",
    "MP2 Energy": "E mp2",
    "MP3 Energy": "E mp3",
    "MP4 Energy": "E mp4",
    "MP4DQ Energy": "E mp4dq",
    "MP4SDQ Energy": "E mp4sdq",
    "MP5 Energy": "E mp5",
    "Mulliken Charges": "Mulliken charges",
    "Number of alpha electrons": "N alpha electrons",
    "Number of beta electrons": "N beta electrons",
    "Number of electrons": "N electrons",
    "Polarizability Derivatives": "polarizability derivatives",
    "Polarizability": "polarizability",
    "RMS Density": "RMS density difference",
    "Real atomic weights": "atomic masses",
    "SCF Energy": "E scf",
    "Total Energy": "energy",
    "Virial Ratio": "virial ratio",
    "atomcharges/apt": "atomic polar tensor charges",
    "atomcharges/mulliken": "mulliken charges",
    "coreelectrons": "core electrons per atom",
    "enthalpy": "H",
    "entropy": "S",
    "freeenergy": "G",
    "metadata/basis_set": "basis set name",
    "metadata/cpu_time": "cpu time",
    "metadata/functional": "density functional",
    "metadata/methods": "methods",
    "metadata/package": "code name",
    "metadata/package_version": "code version",
    "metadata/platform": "platform",
    "metadata/success": "success",
    "metadata/symmetry_detected": "symmetry group",
    "metadata/symmetry_used": "symmetry group used",
    "metadata/wall_time": "elapsed time",
    "moenergies": "MO energies",
    "mosyms": "MO symmetries",
    "nbasis": "NBF",
    "nmo": "NMO",
    "pressure": "P",
    "rotconstants": "rotational constants",
    "scfenergies": "SCF energies",
    "temperature": "T",
    "ts frequency": "transition frequency",
    "vibdisps": "vibrational displacements",
    "vibfconsts": "vibrational force constants",
    "vibfreqs": "vibrational frequencies",
    "vibirs": "vibrational IR intensities",
    "vibramans": "vibrational Raman intensities",
    "vibrmasses": "vibrational reduced masses",
    "vibsyms": "vibrational symmetries",
    "zpve": "ZPE",
}

metadata["results"] = {
    "model": {
        "description": "The model string",
        "dimensionality": "scalar",
        "type": "string",
    },
    "density functional": {
        "description": "The density functional",
        "dimensionality": "scalar",
        "methods": ["DFT"],
        "type": "string",
        "format": "",
    },
    "basis set name": {
        "description": "The basis set used in the calculation",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "DfE0": {
        "calculation": ["thermodynamics"],
        "description": "the energy of formation",
        "dimensionality": "scalar",
        "property": "DfE0#Gaussian#{model}",
        "type": "float",
        "units": "kJ/mol",
        "format": ".2f",
    },
    "DfH0": {
        "calculation": ["thermodynamics"],
        "description": "the enthalpy of formation",
        "dimensionality": "scalar",
        "property": "DfH0#Gaussian#{model}",
        "type": "float",
        "units": "kJ/mol",
        "format": ".2f",
    },
    "energy": {
        "description": "total energy including all terms",
        "dimensionality": "scalar",
        "property": "E#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E 0": {
        "calculation": ["thermodynamics"],
        "description": "the 0 K energy from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "property": "E(0 K)#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "H": {
        "calculation": ["thermodynamics"],
        "description": "The enthalpy of the system",
        "dimensionality": "scalar",
        "property": "H#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "S": {
        "calculation": ["thermodynamics"],
        "description": "The entropy of the system",
        "dimensionality": "scalar",
        "property": "S#Gaussian#{model}",
        "type": "float",
        "units": "E_h/K",
        "format": ".6f",
    },
    "G": {
        "calculation": ["thermodynamics"],
        "description": "The free energy of the system",
        "dimensionality": "scalar",
        "property": "G#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "ZPE": {
        "calculation": ["energy", "thermodynamics"],
        "description": "the ZPE from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "property": "ZPE#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "P": {
        "calculation": ["optimization", "vibrations"],
        "description": "Pressure",
        "dimensionality": "scalar",
        "property": "P#Gaussian#{model}",
        "type": "float",
        "units": "atm",
        "format": ".2f",
    },
    "T": {
        "calculation": ["optimization", "vibrations"],
        "description": "Temperature",
        "dimensionality": "scalar",
        "property": "T#Gaussian#{model}",
        "type": "float",
        "units": "K",
        "format": ".2f",
    },
    "transition state frequency": {
        "calculation": ["optimization", "vibrations"],
        "description": "Transition state curvature",
        "dimensionality": "scalar",
        "property": "transition state frequency#Gaussian#{model}",
        "type": "float",
        "units": "cm^-1",
        "format": ".1f",
    },
    "N saddle modes": {
        "calculation": ["optimization", "vibrations"],
        "description": "Number of saddle point modes",
        "dimensionality": "scalar",
        "property": "number of saddle modes#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "gradients": {
        "calculation": ["energy", "thermodynamics"],
        "description": "the Cartesian gradients",
        "dimensionality": ["natoms", 3],
        "property": "gradients#Gaussian#{model}",
        "type": "float",
        "units": "E_h/a0",
        "format": ".6f",
    },
    "force constants": {
        "calculation": ["thermodynamics"],
        "description": "the Cartesian force constants",
        "dimensionality": "[natoms*(natoms+1)]",
        "property": "force constants#Gaussian#{model}",
        "type": "float",
        "units": "kcal/mol/Å^2",
        "format": ".2f",
    },
    "E scf": {
        "description": "the SCF energy from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "type": "float",
        "property": "E scf#Gaussian#{model}",
        "units": "E_h",
        "format": ".6f",
    },
    "E ccsd_t": {
        "description": "the CCSD(T) energy",
        "dimensionality": "scalar",
        "methods": ["CCSD(T)"],
        "property": "E ccsd_t#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E cc": {
        "description": "the CCD or CCSD energy",
        "dimensionality": "scalar",
        "methods": ["CCD", "CCSD", "CCSD(T)"],
        "property": "E cc#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E qcisd_t": {
        "description": "the CCSD(T) energy",
        "dimensionality": "scalar",
        "methods": ["QCISD(T)"],
        "property": "E qcisd_t#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E qcisd": {
        "description": "the QCISD energy",
        "dimensionality": "scalar",
        "methods": ["QCISD", "QCISD(T)"],
        "property": "E _qcisd#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp2": {
        "description": "the MP2 energy",
        "dimensionality": "scalar",
        "methods": [
            "MP2",
            "MP3",
            "MP4",
            "MP4SDQ",
            "MP5",
            "QCISD",
            "QCISD(T)",
            "CCSD",
            "CCSD(T)",
        ],
        "type": "float",
        "property": "E mp2#Gaussian#{model}",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp3": {
        "description": "the MP3 energy",
        "dimensionality": "scalar",
        "methods": [
            "MP3",
            "MP4",
            "MP4SDQ",
            "MP5",
            "QCISD",
            "QCISD(T)",
            "CCSD",
            "CCSD(T)",
        ],
        "type": "float",
        "property": "E mp3#Gaussian#{model}",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp4": {
        "description": "the MP4 energy",
        "dimensionality": "scalar",
        "methods": ["MP4", "MP4SDQ", "MP5", "CCSD", "CCSD(T)"],
        "property": "E mp4#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp4dq": {
        "description": "the MP4DQ energy",
        "dimensionality": "scalar",
        "methods": ["MP4", "MP4SDQ", "MP5", "QCISD", "QCISD(T)", "CCSD", "CCSD(T)"],
        "property": "E mp4dq#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp4sdq": {
        "description": "the MP4SDQ energy",
        "dimensionality": "scalar",
        "methods": ["MP4", "MP4SDQ", "MP5", "QCISD", "QCISD(T)", "CCSD", "CCSD(T)"],
        "property": "E mp4sdq#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "E mp5": {
        "description": "the MP5 energy",
        "dimensionality": "scalar",
        "methods": ["MP5"],
        "property": "E mp5#Gaussian#{model}",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "symmetry group": {
        "description": "The symmetry of the system",
        "dimensionality": "scalar",
        "property": "symmetry group#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "symmetry group used": {
        "description": "The symmetry used in the calculation",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "virial ratio": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "the virial ratio",
        "dimensionality": "scalar",
        "property": "virial ratio#Gaussian#{model}",
        "methods": ["DFT", "HF", "MP2", "MP3", "MP4"],
        "type": "float",
        "format": ".4f",
    },
    "RMS density difference": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "the RMS density difference in the SCF",
        "dimensionality": "scalar",
        "methods": ["DFT", "HF", "MP2", "MP3", "MP4"],
        "type": "float",
        "format": ".2e",
    },
    "N electrons": {
        "description": "Number of electrons",
        "dimensionality": "scalar",
        "property": "N electrons#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "N alpha electrons": {
        "description": "Number of α electrons",
        "dimensionality": "scalar",
        "property": "N alpha electrons#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "N beta electrons": {
        "description": "Number of β electrons",
        "dimensionality": "scalar",
        "property": "N beta electrons#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "E alpha gap": {
        "description": (
            "The gap between the α-LUMO and α-HOMO or conduction and valence bands"
        ),
        "dimensionality": "scalar",
        "property": "E alpha gap#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E alpha homo": {
        "description": "The energy of highest occupied MO (α-HOMO)",
        "dimensionality": "scalar",
        "property": "E alpha homo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E alpha nhomo": {
        "description": "The energy of the next-to-highest occupied orbital (α-NHOMO)",
        "dimensionality": "scalar",
        "property": "E alpha nhomo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E alpha lumo": {
        "description": "The energy of lowest unoccupied MO (α-LUMO)",
        "dimensionality": "scalar",
        "property": "E alpha lumo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E alpha slumo": {
        "description": "The energy of the second lowest unoccupied orbital (α-SLUMO)",
        "dimensionality": "scalar",
        "property": "E alpha slumo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "alpha HOMO orbital number": {
        "description": "The orbital number of the α-HOMO (starting at 1)",
        "dimensionality": "scalar",
        "type": "integer",
        "format": "",
    },
    "alpha HOMO symmetry": {
        "description": "The symmetry of the highest occupied MO (α-HOMO)",
        "dimensionality": "scalar",
        "property": "alpha HOMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "alpha NHOMO symmetry": {
        "description": "The symmetry of the next-to-highest occupied orbital (α-NHOMO)",
        "dimensionality": "scalar",
        "property": "alpha NHOMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "alpha LUMO symmetry": {
        "description": "The symmetry of the lowest unoccupied MO (α-LUMO)",
        "dimensionality": "scalar",
        "property": "alpha LUMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "alpha SLUMO symmetry": {
        "description": "The symmetry of the second lowest unoccupied orbital (α-SLUMO)",
        "dimensionality": "scalar",
        "property": "alpha SLUMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "E beta gap": {
        "description": (
            "The gap between the β-LUMO and β-HOMO or conduction and valence bands"
        ),
        "dimensionality": "scalar",
        "property": "E beta gap#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E beta homo": {
        "description": "The energy of highest occupied MO (β-HOMO)",
        "dimensionality": "scalar",
        "property": "E beta homo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E beta nhomo": {
        "description": "The energy of the next-to-highest occupied orbital (β-NHOMO)",
        "dimensionality": "scalar",
        "property": "E beta nhomo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E beta lumo": {
        "description": "The energy of lowest unoccupied MO (β-LUMO)",
        "dimensionality": "scalar",
        "property": "E beta lumo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "E beta slumo": {
        "description": "The energy of the second lowest unoccupied orbital (β-SLUMO)",
        "dimensionality": "scalar",
        "property": "E beta slumo#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "beta HOMO orbital number": {
        "description": "The orbital number of the β-HOMO (starting at 1)",
        "dimensionality": "scalar",
        "type": "integer",
        "format": "",
    },
    "beta HOMO symmetry": {
        "description": "The symmetry of the highest occupied MO (β-HOMO)",
        "dimensionality": "scalar",
        "property": "beta HOMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "beta NHOMO symmetry": {
        "description": "The symmetry of the next-to-highest occupied orbital (β-NHOMO)",
        "dimensionality": "scalar",
        "property": "beta NHOMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "beta LUMO symmetry": {
        "description": "The symmetry of the lowest unoccupied MO (β-LUMO)",
        "dimensionality": "scalar",
        "property": "beta LUMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "beta SLUMO symmetry": {
        "description": "The symmetry of the second lowest unoccupied orbital (β-SLUMO)",
        "dimensionality": "scalar",
        "property": "beta SLUMO symmetry#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "multipole reference point": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The reference point for the multipole expansions",
        "dimensionality": "[3]",
        "property": "multipole reference point#Gaussian#{model}",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "dipole moment": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The dipole moment",
        "dimensionality": "[3]",
        "property": "dipole moment vector#Gaussian#{model}",
        "type": "float",
        "units": "debye",
        "format": ".6f",
    },
    "dipole moment magnitude": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The magnitude of the dipole moment",
        "dimensionality": "scalar",
        "property": "dipole moment#Gaussian#{model}",
        "type": "float",
        "units": "debye",
        "format": ".6f",
    },
    "dipole derivatives": {
        "calculation": ["thermodynamics"],
        "description": (
            "The derivatives of the dipole moment with respect to the nuclear "
            "coordinates"
        ),
        "dimensionality": "[natoms, 3]",
        "property": "dipole derivatives#Gaussian#{model}",
        "type": "float",
        "units": "debye/Å",
        "format": ".6f",
    },
    "quadrupole moment": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The quadrupole moment",
        "dimensionality": "[6]",
        "property": "quadrupole moment#Gaussian#{model}",
        "type": "float",
        "units": "debye*Å",
        "format": ".6f",
    },
    "octapole moment": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The octapole moment",
        "dimensionality": "[10]",
        "property": "octapole moment#Gaussian#{model}",
        "type": "float",
        "units": "debye*Å^2",
        "format": ".6f",
    },
    "hexadecapole moment": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The hexadecapole moment",
        "dimensionality": "[15]",
        "property": "hexadecapole moment#Gaussian#{model}",
        "type": "float",
        "units": "debye*Å^3",
        "format": ".6f",
    },
    "polarizability": {
        "calculation": ["thermodynamics"],
        "description": "The polarizability",
        "dimensionality": "[6]",
        "property": "polarizability#Gaussian#{model}",
        "type": "float",
        "units": "a_0^3",
        "format": ".6f",
    },
    "polarizability derivatives": {
        "calculation": ["thermodynamics"],
        "description": "The derivatives of the polarizability",
        "dimensionality": "[natoms, 6]",
        "property": "polarizability derivatives#Gaussian#{model}",
        "type": "float",
        "units": "a_0^2",
        "format": ".6f",
    },
    "hyperpolarizability": {
        "calculation": ["thermodynamics"],
        "description": "The hyperpolarizability",
        "dimensionality": "[10]",
        "property": "hyperpolarizability#Gaussian#{model}",
        "type": "float",
        "units": "a_0^5/e",
        "format": ".6f",
    },
    "NBF": {
        "description": "Number of basis functions",
        "dimensionality": "scalar",
        "property": "NBF#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "NMO": {
        "description": "Number of molecular orbitals",
        "dimensionality": "scalar",
        "property": "NMO#Gaussian#{model}",
        "type": "integer",
        "format": "",
    },
    "SEAMM elapsed time": {
        "description": "total elapsed time for Gaussian",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
        "format": ".1f",
    },
    "SEAMM np": {
        "description": "number of processors for Gaussian",
        "dimensionality": "scalar",
        "type": "integer",
        "format": "d",
    },
    "optimization is converged": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "whether the geometry optimization converged",
        "dimensionality": "scalar",
        "type": "boolean",
        "format": "s",
    },
    "N steps optimization": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "number of optimization steps",
        "dimensionality": "scalar",
        "property": "N steps optimization#Gaussian#{model}",
        "type": "integer",
    },
    "RMS atom force": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the RMS force on the atoms",
        "dimensionality": "scalar",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "RMS atom force threshold": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the RMS force threshold for the atoms",
        "dimensionality": "scalar",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "RMS atom force trajectory": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "trajectory of the RMS force on the atoms",
        "dimensionality": "[N steps optimization]",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "RMS atom displacement": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the RMS displacement of the atoms",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "RMS atom displacement threshold": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the RMS displacement threshold for the atoms",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "RMS atom displacement trajectory": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "trajectory of the RMS displacement of the atoms",
        "dimensionality": "[N steps optimization]",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "maximum atom force": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the maximum force on an atom",
        "dimensionality": "scalar",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "maximum atom force threshold": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the maximum force threshold for an atom",
        "dimensionality": "scalar",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "maximum atom force trajectory": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "trajectory of the maximum force on an atom",
        "dimensionality": "[N steps optimization]",
        "type": "float",
        "units": "E_h/Å",
        "format": ".6f",
    },
    "maximum atom displacement": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the maximum displacement of an atom",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "maximum atom displacement threshold": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "the maximum displacement threshold for an atom",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "maximum atom displacement trajectory": {
        "calculation": ["optimization", "thermodynamics"],
        "description": "trajectory of the maximum displacement of an atom",
        "dimensionality": "[N steps optimization]",
        "type": "float",
        "units": "Å",
        "format": ".6f",
    },
    "mulliken charges": {
        "description": "The Mulliken charges of the atoms",
        "dimensionality": "[natoms]",
        "property": "Mulliken charges#Gaussian#{model}",
        "type": "float",
        "format": ".6f",
    },
    "atomic polar tensor charges": {
        "description": "The atomic polar tensor charges of the atoms",
        "dimensionality": "[natoms]",
        "property": "atomic polar tensor charges#Gaussian#{model}",
        "type": "float",
        "format": ".6f",
    },
    "atomic masses": {
        "description": "The atomic masses of the atoms",
        "dimensionality": "[natoms]",
        "property": "atomic masses#Gaussian#{model}",
        "type": "float",
        "units": "Da",
        "format": "",
    },
    "core electrons per atom": {
        "description": "The number of core electrons per atom",
        "dimensionality": "[natoms]",
        "type": "integer",
        "format": "",
    },
    "cpu time": {
        "description": "The cpu time for the calculation",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "methods": {
        "description": (
            "The method used in each optimization step or step of the calculation"
        ),
        "dimensionality": "[nsteps]",
        "type": "string",
        "format": "",
    },
    "code name": {
        "description": "The name of the code (Gaussian)",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "code version": {
        "description": "The version of the Gaussian code (year+revision)",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "platform": {
        "description": "The platform used for the calculation",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "success": {
        "description": "Whether the calculation was successful",
        "dimensionality": "scalar",
        "type": "boolean",
        "format": "",
    },
    "elapsed time": {
        "description": "The wall-clock time for the calculation",
        "dimensionality": "scalar",
        "type": "string",
        "format": "",
    },
    "MO energies": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The energy of the molecular orbitals (MOs)",
        "dimensionality": "[[nmo]*nspin]",
        "property": "MO energies#Gaussian#{model}",
        "type": "float",
        "units": "eV",
        "format": ".3f",
    },
    "MO symmetries": {
        "calculation": ["energy", "optimization", "thermodynamics"],
        "description": "The symmetry of the MOs",
        "dimensionality": "[nmo]",
        "property": "MO symmetries#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "rotational constants": {
        "calculation": ["energy", "optimization", "vibrations"],
        "description": "Rotational constants",
        "dimensionality": "[[3]*nsteps]",
        "type": "float",
        "units": "GHz",
        "format": ".3f",
    },
    "vibrational displacements": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational displacements",
        "dimensionality": "[nvibs, natoms, 3]",
        "type": "float",
        "format": ".2f",
    },
    "vibrational frequencies": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational frequencies",
        "dimensionality": "[nvibs]",
        "property": "vibrational frequencies#Gaussian#{model}",
        "type": "float",
        "units": "cm^-1",
        "format": ".1f",
    },
    "vibrational IR intensities": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational IR intensities",
        "dimensionality": "[nvibs]",
        "property": "vibrational IR intensities#Gaussian#{model}",
        "type": "float",
        "units": "km/mol",
        "format": ".4f",
    },
    "vibrational Raman intensities": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational Raman intensities",
        "dimensionality": "[nvibs]",
        "property": "vibrational Raman intensities#Gaussian#{model}",
        "type": "float",
        "units": "Å^4/Da",
        "format": ".4f",
    },
    "vibrational reduced masses": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational reduced masses",
        "dimensionality": "[nvibs]",
        "type": "float",
        "units": "Da",
        "format": ".4f",
    },
    "vibrational symmetries": {
        "calculation": ["optimization", "vibrations"],
        "description": "Vibrational symmetries",
        "dimensionality": "[nvibs]",
        "property": "vibrational symmetries#Gaussian#{model}",
        "type": "string",
        "format": "",
    },
    "E thermal": {
        "calculation": ["thermodynamics"],
        "description": "the thermal energy from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "deltaE mp2": {
        "calculation": ["energy", "optimization"],
        "description": "dE from MP2 from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "deltaE cbs": {
        "calculation": ["energy", "optimization"],
        "description": "dE from the basis extrapolation from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "deltaE mp34": {
        "calculation": ["energy", "optimization"],
        "description": "dE of MP3-4 from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E ccsd": {
        "calculation": ["energy", "optimization"],
        "description": "dE of CCSD from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": ["CBS-4M", "CBS-QB3", "CBS-APNO"],
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E internals": {
        "calculation": ["energy", "optimization"],
        "description": "dE from internals from the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E empirical": {
        "calculation": ["energy", "optimization"],
        "description": "Empirical correction to the CBS/Gn extrapolation",
        "dimensionality": "scalar",
        "methods": composite,
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E diffuse": {
        "calculation": ["energy", "optimization"],
        "description": "Plus term for the Gn extrapolation",
        "dimensionality": "scalar",
        "methods": ["G1", "G2", "G3", "G4"],
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E 2df": {
        "calculation": ["energy", "optimization"],
        "description": "The delta energy for 2D's and F term for the Gn extrapolation",
        "dimensionality": "scalar",
        "methods": ["G1", "G2", "G3", "G4"],
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E g3xp": {
        "calculation": ["energy", "optimization"],
        "description": "Delta G3XP term for the Gn extrapolation",
        "dimensionality": "scalar",
        "methods": ["G1", "G2", "G3", "G4"],
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "delta E hf": {
        "calculation": ["energy", "optimization"],
        "description": "delta E(HF) term for the Gn extrapolation",
        "dimensionality": "scalar",
        "methods": ["G1", "G2", "G3", "G4"],
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
    "Wiberg bond order matrix": {
        "description": "the Wiberg bond order matrix",
        "dimensionality": "[natoms * [natoms]]",
        "type": "float",
        "format": ".3f",
    },
    "SCF energies": {
        "description": "The energies per step",
        "dimensionality": "[nsteps]",
        "type": "float",
        "units": "E_h",
        "format": ".6f",
    },
}



# -*- coding: utf-8 -*-

"""The graphical part of a Thermodynamics step"""

import pprint  # noqa: F401
import tkinter as tk
import tkinter.ttk as ttk

import gaussian_step  # noqa: F401, E999
from seamm_util import ureg, Q_, units_class  # noqa: F401, E999
import seamm_widgets as sw


class TkThermodynamics(gaussian_step.TkOptimization):
    """
    The graphical part of a Thermodynamics step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Thermodynamics_parameters.py

    See Also
    --------
    Thermodynamics, TkThermodynamics,
    ThermodynamicsParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
        )

    def create_dialog(self, title="Edit Gaussian Thermodynamics Step"):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the Thermodynamics_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkThermodynamics.reset_dialog
        """

        # Let parent classes do their thing.
        super().create_dialog(title=title)

        # Shortcut for parameters
        P = self.node.parameters

        # Frame to isolate widgets
        thermo_frame = self["thermodynamics"] = ttk.LabelFrame(
            self["frame"],
            borderwidth=4,
            relief="sunken",
            text="Thermodynamics",
            labelanchor="n",
            padding=10,
        )

        for key in gaussian_step.ThermodynamicsParameters.parameters:
            self[key] = P[key].widget(thermo_frame)

        # bindings...
        for key in ("optimize first",):
            self[key].bind("<<ComboboxSelected>>", self.reset_dialog)
            self[key].bind("<Return>", self.reset_dialog)
            self[key].bind("<FocusOut>", self.reset_dialog)

        # Top level needs to call reset_dialog
        if self.node.calculation == "thermodynamics":
            self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets, letting our parents go first."""
        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        optimize_first = self["optimize first"].get() != "no"

        row = 0
        # Whether to just write input
        self["input only"].grid(row=row, column=0, sticky=tk.W)
        row += 1

        self["thermodynamics"].grid(row=row, column=0)
        self.reset_thermodynamics()

        self["calculation"].grid(row=row, column=1)
        self.reset_calculation()
        row += 1

        if optimize_first:
            self["optimization"].grid(row=row, column=0)
            self.reset_optimization()

        self["convergence frame"].grid(row=row, column=1)
        self.reset_convergence()
        row += 1

        self["structure frame"].grid(row=row, column=0, columnspan=2)

        return row

    def reset_thermodynamics(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Thermodynamics_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkThermodynamics.create_dialog
        """

        # Remove any widgets previously packed
        frame = self["thermodynamics"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as "method" here
        row = 0
        widgets = []
        for key in ("optimize first",):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        # Align the labels
        sw.align_labels(widgets, sticky=tk.E)

    def right_click(self, event):
        """
        Handles the right click event on the node.

        Parameters
        ----------
        event : Tk Event

        Returns
        -------
        None

        See Also
        --------
        TkThermodynamics.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)
